;*************************************************************************
;*									 *
;*				AVR universal Bootloader		 *
;*									 *
;*			Author: Peter Dannegger				 *
;*									 *
;*************************************************************************
.include "tinyload.h"
.listmac
.list
	.org	0x0000
#ifdef	USE_JMP
	jmp	init
#else
	rjmp	init
#endif
	.org	BootStart
init:
	cli
	ldi	a0, low (RamEnd)
	out	SPL, a0
#ifdef USE_SPH
	ldi	a0, high(RamEnd)
	out	SPH, a0
#endif
	clr	zerol
	clr	zeroh
	clr	Flags
	sbi	SRX_PORT, SRX
	sbi	STX_PORT, STX
	sbi	STX_DDR, STX
;-------------------------------------------------------------------------
.include "abaud.inc"
;-------------------------------------------------------------------------
;				Check password
;-------------------------------------------------------------------------
checkpwd:
	ldi	zl, low (2*Password)
	ldi	zh, high(2*Password)
	rjmp	_cpw3
_cpw2:
	rcall	getchar
	cpse	r0, a0
	rjmp	checkpwd
_cpw3:
	lpm	r0, z+
	tst	r0
	brne	_cpw2
;-------------------------------------------------------------------------
;				Receive commands
;-------------------------------------------------------------------------
;A5,00 get bootloader revision
;A5,01 get buffer size
;A5,02 get target signature
;A5,03 get user flash size
;A5,04 program flash
;A5,05 start application
;-------------------------------------------------------------------------
mainloop:
	ldi	a0, SUCCESS
	sbrc	Flags, Ferror
	ldi	a0, FAIL
	cbr	Flags, 1<<Ferror
	rcall	putchar
_cex1:
	rcall	getchar
	cpi	a0, COMMAND
	brne	_cex1
	rcall	getchar
	cpi	a0, 4
	breq	program			; 4
	brcs	sendmessage		; 0 ... 3
	cpi	a0, 6
	brcc	_cex2			; >5
_timeout:
	out	STX_PORT, zerol
	out	STX_DDR, zerol
	out	SRX_PORT, zerol
#ifdef USE_JMP
	jmp	0x0000
#else
#ifdef BootStart1
	rjmp	0x0000			; run application
#else
	rjmp	BootStart-1		; run user application
#endif
#endif
_cex2:
	sbr	flags, 1<<Ferror
	rjmp	mainloop
;-------------------------------------------------------------------------
;				Send Messages
;-------------------------------------------------------------------------
;input: a0 = Message index
;
SendMessage:
	mov	a2, a0
	ldi	zl, low (2*Messages)
	ldi	zh, high(2*Messages)
	ldi	a0, ANSWER
	rcall	putchar
	ldi	a0, 0
_sme1:
	add	zl, a0
	adc	zh, zeroh
	lpm	a0, z+
	subi	a2, 1
	brcc	_sme1
	mov	a2, a0
_sme2:
	rcall	putchar
	lpm	a0, z+
	subi	a2, 1
	brcc	_sme2
	rjmp	mainloop
;-------------------------------------------------------------------------
#ifdef BootStart1
.include "program2.inc"		; mega with bootstart fuse set
#else
.include "program.inc"		; tiny, mega without RWW section
#endif
	rjmp	MainLoop
;-------------------------------------------------------------------------
.include "uart.inc"
;-------------------------------------------------------------------------
Password:
	.db	"Peda", 0, 0
Messages:
	.db \
	2, high(Version), low(Version), \
	2, high(2*BufferSize), low(2*BufferSize), \
	3, SIGNATURE_000, SIGNATURE_001, SIGNATURE_002, \
	3, byte3(UserFlash), byte2(UserFlash), byte1(UserFlash)
;-------------------------------------------------------------------------
